"""Any option mentioning "a line" means either a number or a regexp"""

import sys
from collections import defaultdict
from functools import partial

from pysyte import __version__
from pysyte.cli.arguments import ArgumentsParser
from pysyte.types import lines as pylines
from pysyte.bash.screen import alt_screen


class LinesParser(ArgumentsParser):
    """Add option sets to the parser"""
    def __init__(self, parser_):
        super().__init__(parser_)
        self.version = __version__
        self.groups = defaultdict(list)

    def add_files(self, name=None, action=None):
        name_ = name if name else 'files'
        action_ = action if action else 'show'
        self.positional(name_, help=f'{name_} to {action_}')
        return self

    def add_to(self, letter, name, group, type_, help_, default=None):
        letter_ = f'-{letter.lstrip("-")}'
        name_ = f'--{name.lstrip("-")}'
        safe_default = (
            default if default
            else 0 if str(type_).startswith('int')
            else '')
        method = getattr(self, type_, f"not a method for {letter},{name}")
        assert callable(method), f'self.{type_}() is {method}'
        if type_ == 'boolean':
            arg = method(letter_, name_, help=help_)
        else:
            arg = method(letter_, name_, help=help_, default=safe_default)
        self.groups[group].append(arg)

    def add_lines(self):
     args = {
      'a': ['at', 'lines', 'inty', 'show that line'],
      'c': ['copy', 'clipboard', 'boolean', 'copy text to clipboard'],
      # 'd': ['delete', 'lines', 'string', 'lines to be deleted'],
      'e': ['editor', 'ed', 'string', 'edit with, e.g., vim'],
      'f': ['first', 'lines', 'inty', 'the first line to show', '1'],
      'i': ['stdin', 'stdin', 'boolean', '(aka -) wait for text from stdin'],
      'l': ['last', 'lines', 'inty', 'the last line to show', '0'],
      'n': ['numbers', 'lines', 'boolean', 'show line numbers'],
      'p': ['paste', 'clipboard', 'boolean', 'paste text from clipboard'],
      's': ['expression', 'ed', 'string', 'sed expression to be executed'],
      'v': ['remove', 'lines', 'inty', 'remove that line'],
      'V': ['version', 'version', 'boolean', 'show version'],
      'w': ['width', 'lines', 'integer', 'max width of lines', '80'],
     }
     [self.add_to(letter, *args) for letter, args in args.items()]
     return self

    def post_parser(self, args):
        if args.version and self.version:
            sys.stdout.write(f'{sys.argv[0]} version: {self.version}')
            raise SystemExit
        args.sed = partial(
            pylines.reformat_lines, numbers=args.numbers, width=args.width)
        args.alt_screen = alt_screen()
        return args


def add_args(old_parser):
    """Create a new parser to handle some lines from given parser"""
    result = LinesParser(old_parser.parser)
    result.add_lines()
    return result



"""Handle XDG config files as yaml"""

from functools import partial

from pysyte.config.types import YamlConfiguration
from pysyte.config.types import YamlyConfiguration
from pysyte.config.types import IniConfiguration
from pysyte.iteration import first
from pysyte.oss.linux import xdg_home
from pysyte.types.paths import path


def home_config(string):
    return first_config(string, [xdg_home()])


def any_config_type(string):
    for path_, type_ in config_types(string):
        if path_.isfile():
            return type_(path_)
    return {}


def first_config(string, paths):
    return first(any_config_type(_ / string) for _ in paths)


def any_config(string):
    return first_config(string, (xdg_home(), path('/etc')))


def all_configs(string):
    return (any_config_type(_ / string) for _ in (path('/etc'), xdg_home()) if _)


def etc_config(string):
    return any_config_type(path('/etc') / string)


def config_types(stem):
    config_exts = (
        ('yml', YamlConfiguration),
        ('yaml', YamlConfiguration),
        ('yamly', YamlyConfiguration),
        ('ini', IniConfiguration),
    )
    return [(stem.add_missing_ext(e), t) for e, t in config_exts]



"""Handle datetime for pysyte types"""

import datetime
epoch = datetime.datetime.fromtimestamp(0)


def taken(diff):
    """Convert a time diff to total number of microseconds"""
    microseconds = diff.seconds * 1_000 + diff.microseconds
    return abs(diff.days * 24 * 60 * 60 * 1_000 + microseconds)


def seconds_taken(diff):
    """Convert a time diff to total number of seconds"""
    return taken(diff) // 1_000


def now():
    """Convenience method to call datetime's now"""
    return datetime.datetime.now()


def microseconds_since_epoch():
    """Number of microseconds since the epoch"""
    return taken(now() - epoch)


def seconds_since_epoch():
    """Number of seconds since the epoch"""
    return seconds_taken(now() - epoch)



"""Five most popular similarity measures

You won't believe #5, said https://dataaspirant.com/2015/04/11/five-most-popular-similarity-measures-implementation-in-python/

Similarity:

The similarity measure is the measure of how much alike two data objects are. Similarity measure in a data mining context is a distance with dimensions representing features of the objects. If this distance is small, it will be the high degree of similarity where large distance will be the low degree of similarity.

The similarity is subjective and is highly dependent on the domain and application. For example, two fruits are similar because of color or size or taste. Care should be taken when calculating distance across dimensions/features that are unrelated. The relative values of each element must be normalized, or one feature could end up dominating the distance calculation. Similarity are measured in the range 0 to 1 [0,1].

"""

from math import sqrt
from decimal import Decimal

def similar(x, y):
    """Two main consideration about similarity:

    Similarity = 1 if X = Y         (Where X, Y are two objects)
    Similarity = 0 if X ≠ Y
    """
    try:
        setattr(x, 'similar', lambda z: similar (x, z))
        setattr(y, 'similar', lambda z: similar (y, z))
    except AttributeError:
        pass
    return x == y


def euclidean(x, y):
    """Euclidean distance is the most common use of distance.

    In most cases when people said about distance, they will refer to Euclidean distance. Euclidean distance is also known as simply distance. When data is dense or continuous, this is the best proximity measure.

    The Euclidean distance between two points is the length of the path connecting them.The Pythagorean theorem gives this distance between two points.
    """
    return sqrt(sum(pow(a-b,2) for a, b in zip(x, y)))


def manhattan(x, y):
    """Manhattan distance is a metric in which the distance between two points is the sum of the absolute differences of their Cartesian coordinates

     In a simple way of saying it is the total suzm of the difference between the x-coordinates  and y-coordinates.

     Suppose we have two points A and B if we want to find the Manhattan distance between them, just we have, to sum up, the absolute x-axis and y – axis variation means we have to find how these two points A and B are varying in X-axis and Y- axis. In a more mathematical way of saying Manhattan distance between two points measured along axes at right angles.

     In a plane with p1 at (x1, y1) and p2 at (x2, y2).

     Manhattan distance = |x1 – x2| + |y1 – y2|

     This Manhattan distance metric is also known as Manhattan length, rectilinear distance, L1 distance or L1 norm, city block distance, Minkowski’s L1 distance, taxi-cab metric, or city block distance.
     """
    return sum(abs(a-b) for a,b in zip(x,y))


def minkowski(x, y, p_value):
    """Minkowski distance is a generalized metric form of Euclidean distance and Manhattan distance
    In the equation, d^MKD is the Minkowski distance between the data record i and j, k the index of a variable, n the total number of variables y and λ the order of the Minkowski metric. Although it is defined for any λ > 0, it is rarely used for values other than 1, 2 and ∞.

    The way distances are measured by the Minkowski metric of different orders between two objects with three variables ( In the image it displayed in a coordinate system with x, y ,z-axes).

    Synonyms of Minkowski:
    Different names for the Minkowski distance or Minkowski metric arise from the order:

    λ = 1 is the Manhattan distance. Synonyms are L1-Norm, Taxicab or City-Block distance. For two vectors of ranked ordinal variables, the Manhattan distance is sometimes called Foot-ruler distance.
    λ = 2 is the Euclidean distance. Synonyms are L2-Norm or Ruler distance. For two vectors of ranked ordinal variables, the Euclidean distance is sometimes called Spear-man distance.
    λ = ∞ is the Chebyshev distance. Synonyms are Lmax-Norm or Chessboard distance.
    """
    def nth_root(value, n_root):
        root_value = 1/float(n_root)
        return float(round (Decimal(value) ** Decimal(root_value),3))

    return nth_root(sum(pow(abs(a-b),p_value) for a,b in zip(x, y)),p_value)



"""Linux-specific code"""

import os

from pysyte.types import paths


def xdg_home():
    """path to $XDG_CONFIG_HOME

    >>> assert xdg_home() == os.path.expanduser('~/.config')
    """
    path_to_xdg = os.getenv('XDG_CONFIG_HOME', str(paths.home() / '.config'))
    return paths.path(path_to_xdg)


def xdg_home_config(filename):
    """path to that file in $XDG_CONFIG_HOME

    >>> assert xdg_home_config('fred') == os.path.expanduser('~/.config/fred')
    """
    return xdg_home() / filename


bash_paste = 'xclip -selection clipboard'
bash_copy = 'xclip -selection clipboard -o'



"""This module handles numbers"""

class NAN(ValueError):
    def __init__(self, value):
        super().__init__(f'NAN: {value}')


def _eval(value, kind, default=None):
    try:
        return kind(value)
    except (ValueError, TypeError):
        # if default is None:
            # raise NAN(value)
        if isinstance(value, (str, type(None))):
            return default
        try:
            return len(value)
        except TypeError:
            raise NAN(value)


as_int = lambda x: _eval(x, int)
as_float = lambda x: _eval(x, float)

inty = lambda x: _eval(x, int, x)
floaty = lambda x: _eval(x, float, x)



import atexit

from pysyte.bash.shell import run

_alt_screen_started = False

def _stop_alt_screen():
    if _alt_screen_started:
        run('tput rmcup')


def _start_alt_screen():
    _alt_screen_started = True
    run('tput smcup')


def get_alt_screens():
    return _start_alt_screen, _stop_alt_screen


def alt_screen():
    atexit.register(_stop_alt_screen)
    return _start_alt_screen



"""A module to provide decorators which change methods"""

import pdb

try:
    import pudb as pdb
except ImportError:
    pass


def _represent_arguments(*arguments, **keyword_arguments):
    """Represent the aruments in a form suitable as a key (hashable)

    And which will be recognisable to user in error messages
    >>> print(_represent_arguments([1, 2], **{'fred':'here'}))
    [1, 2], fred='here'
    """
    argument_strings = [repr(a) for a in arguments]
    keyword_strings = [
        '='.join((k, repr(v))) for k, v in keyword_arguments.items()]
    return ', '.join(argument_strings + keyword_strings)


def memoize(method):
    """A new method which acts like the given method but memoizes arguments

    See https://en.wikipedia.org/wiki/Memoization for the general idea
    >>> @memoize
    ... def test(arg):
    ...     print('called')
    ...     return arg + 1
    >>> test(1)
    called
    2
    >>> test(2)
    called
    3
    >>> test(1)
    2

    The returned method also has an attached method "invalidate"
        which removes given values from the cache
        Or empties the cache if no values are given
    >>> test.invalidate(2)
    >>> test(1)
    2
    >>> test(2)
    called
    3
    """
    method.cache = {}

    def invalidate(*arguments, **keyword_arguments):
        key = _represent_arguments(*arguments, **keyword_arguments)
        if not key:
            method.cache = {}
        elif key in method.cache:
            del method.cache[key]
        else:
            raise KeyError(f'Not prevously cached: {method.__name__}({key})')

    def new_method(*arguments, **keyword_arguments):
        """Cache the arguments and return values of the call

        The key cached is the repr() of arguments
            This allows more types of values to be used as keys to the cache
            Such as lists and tuples
        """
        key = _represent_arguments(*arguments, **keyword_arguments)
        if key not in method.cache:
            method.cache[key] = method(*arguments, **keyword_arguments)
        return method.cache[key]
    new_method.invalidate = invalidate
    new_method.__doc__ = method.__doc__
    new_method.__name__ = f'memoize({method.__name__})'
    return new_method


def debug(method):
    """Decorator to debug the given method"""
    def new_method(*args, **kwargs):
        try:
            pdb.runcall(method, *args, **kwargs)
        except pdb.bdb.BdbQuit:
            pass
    new_method.__doc__ = method.__doc__
    new_method.__name__ = f'debug({method.__name__})'
    return new_method



"""Module to handle streams of text from cli arguments"""

import os
import sys
import contextlib
from itertools import chain

from six import StringIO

from pysyte import iteration
from pysyte.cli import arguments
from pysyte.oss.platforms import get_clipboard_data

def parse_args():
    """Parse out command line arguments"""
    parser = arguments.parser(__doc__)
    parser.args('streams', help='streams to use')
    parser.opt('-p', '--paste', 'paste text from clipboard')
    parser.opt('-i', '--stdin', 'wait for text from stdin')
    return parser.parse_args()


def args(parsed_args, name=None, files_only=False):
    """Interpret parsed args to streams"""
    strings = parsed_args.get_strings(name)
    files = [s for s in strings if os.path.isfile(s)]
    if files:
        streams = [open(f) for f in files]
    elif files_only:
        return []
    else:
        streams = []
    if '-' in files or not files or getattr(parsed_args, 'stdin', False):
        streams.append(sys.stdin)
    if getattr(parsed_args, 'paste', not files):
        streams.append(clipboard_stream())
    return streams


def files(parsed_args, name=None):
    return args(parsed_args, name, True)


def all():
    yielded = False
    for path in _arg_files():
        yield open(path)
        yielded = True
    if not yielded or '-' in sys.argv:
        yield sys.stdin


def some():
    if sys.argv[1:]:
        assert _arg_files()
    return any()


def clipboard_stream(name=None):
    stream = StringIO(get_clipboard_data())
    stream.name = name or '<clipboard>'
    return stream


def _arg_files():
    return [a for a in sys.argv[1:] if os.path.isfile(a)]


def _arg_streams():
    """yield streams to all arg.isfile()"""
    for path in _arg_files():
        yield open(path)


def _any():
    try:
        stream = iteration.first(_arg_streams())
        if stream:
            return _arg_streams()
    except ValueError:
        return iter([clipboard_stream(), sys.stdin])



"""Handle different OS platforms"""
import platform as python_platform
import importlib
from subprocess import run

from pysyte import oss
from pysyte.bash import shell

name = python_platform.system().lower()
platform = importlib.import_module(f'pysyte.oss.{name}', oss)


def put_clipboard_data(data):
    run(platform.bash_copy, encoding='utf-8', input=data)


def get_clipboard_data():
    result = run(platform.bash_paste, capture_output=True, encoding='utf-8')
    return result.stdout
